import time

import jax
import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np

from dynax import Flow, DynamicStateFeedbackSystem, ControlAffine
from dynax.linearize import relative_degree, input_output_linearize

class NonlinearDrag(ControlAffine):
  m: float
  r: float
  r2: float
  k: float
  n_states = 2
  n_inputs = 1
  n_outputs = 1

  def f(self, x, u=None, t=None):
    x1, x2 = x
    return jnp.array([
      x2,
      (-self.r * x2 - self.r2 * jnp.abs(x2) * x2 - self.k * x1) / self.m])

  def g(self, x, u=None, t=None):
    return jnp.array([0.0, 1.0 / self.m])

dyn = NonlinearDrag(m=1., r=1., r2=0.2, k=1.)
init_state = np.zeros(dyn.n_states)

# design input signal
T = 100
sr = 100
t = np.arange(int(T*sr))/sr
r = 10*np.sin(2*np.pi*t*0.1)

ref_dyn = dyn.linearize()

fig, ax = plt.subplots(nrows=2)
ax[0].plot(r, label='$r$')
us = []

for output in [0, 1]:
  # estimate relative degree
  x_test = np.random.normal(size=(1000, 2))
  reldeg = relative_degree(dyn, x_test, output=output)
  print("Relative degree:", reldeg)

  # construct input-output linearized model
  feedbacklaw = input_output_linearize(dyn, reldeg, ref=ref_dyn, output=output)
  feedbacksys = DynamicStateFeedbackSystem(dyn, ref_dyn, feedbacklaw)

  # combine model with ODE solver
  feedbackmodel = jax.jit(Flow(feedbacksys).__call__)
  init_xz = jnp.concatenate((init_state, init_state))

  # measure compile time
  start = time.time()
  xz, y_comp = feedbackmodel(init_xz, t, r)
  print(f"Compile+run: {time.time()-start}s")

  # measure jitted run time
  start = time.time()
  for i in range(5):
    xz, y_comp = feedbackmodel(init_xz, t, r)
  print(f"jitted run: {(time.time()-start)/5}s")

  # recompute u from states and input
  x, z = xz[:, :dyn.n_states], xz[:, dyn.n_states:]
  u = jax.vmap(feedbacklaw)(x, z, r)

  us.append(u)

  ax[0].plot(u, label=f'$u$ for $y=x[{output}]$')
  ax[1].plot(x[:, 0], label=f'$x$ for $y=x[{output}]$')
  ax[1].plot(z[:, 0], label=f'$z$ for $y=x[{output}]$')

for a in ax:
  a.legend()

plt.show()

np.all(us[0] == us[1])  # -> True
